//                  MailListSort.java

//              Bubble Sort and Record Sort Demo
//  Alphabetizes and Prints a Mailing List Organized as an Array of Records
//               A Java Screen Based Application

//       Original Program in Pascal by Charles W. Neville, 1982
//    Translated from the Pascal into Java by Charles W. Neville, 1998
//             Copyright Charles W. Neville, 1998                 

//  Long long ago on a machine far far away, I wrote the original version of 
//  this program in Pascal and ran it.  This is a close to literal translation 
//  of the Pascal version into Java, except that the original NESTED record 
//  structure has been flattened.  It does not run, because of the $DATA 
//  statement at the end, a holdover from a batch processing system on an IBM 
//  370, though it will run if that is taken out.

import java.io.*;
import LinearlyOrdered;
import ObjectBubSort;

// member record
class Member implements LinearlyOrdered {

   // member's name
   String last;
   String first;
   char   middle;

   // member's address
   String street;
   String city;
   String state;
   int    zip;

   // IO methods

   // read a member record from BufferedReader br
   // returns the record read -- returns null if there are no more records,
   // throws IOException
   // make it static so we can call it to create a Member object from a file
   public static Member readMember(BufferedReader br) throws IOException {

      Member member = new Member();

      // read member's name
      member.first = br.readLine();  // returns a string -- to end of line
      if (member.first == null) return null;  //end of file -- return null 
      member.middle = (char)br.read();     // returns a single char
      br.readLine();               // get to next line
      member.last = br.readLine();

      // read member's address
      member.street = br.readLine();
      member.city = br.readLine();
      member.state = br.readLine();
      member.zip = Integer.parseInt(br.readLine());  

      return member;

   } // readMember

   // write a member record to PrintWriter pw
   // returns true if all went well, false otherwise
   // does not throw any Exceptions
   public boolean writeMember(PrintWriter pw) {

        // do the writing
        pw.println(first);
        pw.println(middle);
        pw.println(last);
        pw.println(street);
        pw.println(city);
        pw.println(state);
        pw.println(Integer.toString(zip));

        return !pw.checkError();  // unlike other streams, PrintError never 
                // throws an IOException.  So we have to use the checkError() 
                // method to see if an error has occured.  checkError returns 
                // true if an error has occured and false otherwise.  Thus we 
                // return !pw.checkError() so writeMember will return true if 
                // all went well and false otherwise.  
                // checkError also flushes the stream, so we don't need to call 
                // the flush method.                          

   } // writeMember

   // write a member record to screen -- note use of polymorphism
   public void writeMember() {

      System.out.println(first);
      System.out.println(middle);
      System.out.println(last);
      System.out.println(city);
      System.out.println(state);
      System.out.println(Integer.toString(zip));

   } // writeMember

   // interface LinearlyOrdered implementation -- methods gt, lt and eq.

   // is this Member record > Member record b?  Compare alphabetically, with 
   // last name first, then first name, then middle initial, then street, then 
   // city, and finally zip.  gt(b) returns true if this record  > record b, 
   // false otherwise.  Java silliness requires us to parametrize gt as 
   // gt(LinearlyOrdered b) rather than gt(Member b).
   public boolean gt(LinearlyOrdered b) {

      Member c = (Member) b;  // more Java silliness, cast parameter b 
                              // to a Member object

      // compare member's names
      if (this.last.compareTo(c.last) > 0) return true;  // this says,
           // if( this.last > c.last) return true, but have to use compareTo
           // method in place of > operator.
      if (this.last.compareTo(c.last) < 0) return false;  // this says,
           // if( this.last < c.last) return false, but have to use compareTo
           // method in place of < operator.
      // if we get to here, last names are the same, so compare first names.
      if (this.first.compareTo(c.first) > 0) return true;
      if (this.first.compareTo(c.first) < 0) return false;
      // if we get to here, last and first names are the same, so compare 
      // middle initials.
      if (this.middle > c.middle) return true;
      if (this.middle < c.middle) return false;

      // compare member's addresses
      if (this.street.compareTo(c.street) > 0) return true;
      if (this.street.compareTo(c.street) < 0) return false;
      // if we get to here, streets are the same, so compare cities.
      if (this.city.compareTo(c.city) > 0) return true;
      if (this.city.compareTo(c.city) < 0) return false;
      // if we get to here, streets and cities are the same, so compare zips.
      if (this.zip > c.zip) return true;
      if (this.zip < c.zip) return false;
      
      return false;  // two records are equal if we get to here.

   }  // gt

   // is this Member record < Member record b?  Compare alphabetically, with 
   // last name first, then first name, then middle initial, then street, then 
   // city, and finally zip.  gt(b) returns true if this record  < record b, 
   // false otherwise.  Java silliness requires us to parametrize gt as 
   // gt(LinearlyOrdered b) rather than gt(Member b).
   public boolean lt(LinearlyOrdered b) {

      Member c = (Member) b;  // more Java silliness, cast parameter b 
                              // to a Member object

      // compare member's names
      if (this.last.compareTo(c.last) < 0) return true;  // this says,
           // if( this.last < c.last) return true, but have to use compareTo
           // method in place of < operator.
      if (this.last.compareTo(c.last) > 0) return false;  // this says,
           // if( this.last > c.last) return false, but have to use compareTo
           // method in place of > operator.
      // if we get to here, last names are the same, so compare first names.
      if (this.first.compareTo(c.first) < 0) return true;
      if (this.first.compareTo(c.first) > 0) return false;
      // if we get to here, last and first names are the same, so compare 
      // middle initials.
      if (this.middle < c.middle) return true;
      if (this.middle > c.middle) return false;

      // compare member's addresses
      if (this.street.compareTo(c.street) < 0) return true;
      if (this.street.compareTo(c.street) > 0) return false;
      // if we get to here, streets are the same, so compare cities.
      if (this.city.compareTo(c.city) < 0) return true;
      if (this.city.compareTo(c.city) > 0) return false;
      // if we get to here, streets and cities are the same, so compare zips.
      if (this.zip < c.zip) return true;
      if (this.zip > c.zip) return false;
      
      return false;  // two records are equal if we get to here.

   }  // lt

   // is this Member record = Member record b?  eq(b) returns true if 
   // this record  = record b, false otherwise.  Java silliness requires us to 
   // parametrize eq as eq(LinearlyOrdered b) rather than eq(Member b).
   public boolean eq(LinearlyOrdered b) {

      return !gt(b) && !lt(b);

   } // eq 

}  // class Member
     
// read list of member records from file1 into an array L, bubble sort the 
// list, and write the sorted array out to file2.  This program gets its file
// arguments from the command line, as follows:
//
//    java MailListSort file1 file2   read unsorted list from file1
//                                    write sorted list to file2
//
public class MailListSort {

   private final String[] usageMessage = {
      "Usage: java MailListSort file1 file2   read unsorted list from file1",
      "                                       write sorted list to file2"};

   private final int numMax = 200;  // maximum number of members 

   private Member[] L;  // list of members.  Capital L because it is an array.

    // various utility methods

    // print an array of strings, such as the usageMessage
    public void printStringArray(String s[]) {

       for(int i = 0; i < s.length; i++) {
          System.out.println(s[i]);
       }
    } // printStringArray

    // print usage message
    public void printUsageMessage() {

       printStringArray(usageMessage);
    } // printUsageMessage

    // print security error message
    public void printSecurityError() {

       System.out.println("Insufficient privileges for attempted operation");
    } // printSecurityError

    // print file not found error message
    public void printFileNotFoundError() {

       System.out.println("Input file not found");
    }

    // IO methods

    // read unsorted mailing list into array L from file1
    // returns true if all went well, false otherwise
    // side effects: upon exit from the method,
    //    array L has length = the number of records read.
    //    array L contains the list of records read from file1.
    //    if file1 contains more than numMax records, only the first numMax
    //       records will be read into L.
    //    if the method encounters an Exception, array L will = null.
    public boolean readInList(String file1) {

        BufferedReader br;     // read from br

        int numRead;  // number of records actually read

        Member[] M = new Member[numMax];  // temporary array to hold list of
                        // records.  We need M so we can later make the list 
                        // array L have length exactly equal to the number of 
                        // records read.

        // BufferedReader constructors don't throw Exceptions, but constructing 
        // a FileReader object throws a FileNotFoundException (because the 
        // files must be  opened), so we have to use try -- catch clauses to 
        // catch the exceptions.
        try {
           br = new BufferedReader(new FileReader(file1) );  // this creates
               // a BufferedReader attached to a FileReader and the physical 
               // file named file1, and it opens file1 for reading.
        }
        catch (FileNotFoundException e) {
            printFileNotFoundError();
            return false;  // readIn operation failed
        }

        // do the actual reading into temporary array M.  Since the Member 
        // readMember method throws an IOException, we have to use a try -- 
        // catch clause again.
        try {
           // read member records into array L up to end of file
           // buffered IO for efficiency
           numRead = 0;
           while((numRead < M.length) && 
                  ((M[numRead] = Member.readMember(br)) != null)) {

               // echoprint record just read
               // for debugging -- uncomment next line if program fails
               // M[numRead].writeMember();

               numRead++;
           }
        }
        catch (NumberFormatException e) {
            System.out.println("Invalid zip code");
            return false;  // readInList operation failed
        }
        catch(IOException e) {
           System.out.println("Error reading from file");
           return false;  // readInList operation failed
        }

        // construct list array L with length exactly equal to the number of 
        // records read, and read member list from array M into array L.
        L = new Member[numRead];
        for(int i = 0; i < numRead; i++) {
            L[i] = M[i];
        }

        return true;  // readInList operation succeeded

    }  // readInList

    // write sorted mailing list out to file2
    // returns true if all went well, false otherwise
    public boolean writeOutList(String file2) {

        PrintWriter pw;
        boolean noError;
        int i;  // loop counter

        // PrintWriter methods and BufferedWriter constructors don't throw 
        // IOExceptions, but constructing a FileWriter object throws a 
        // SecurityException and an IOException (because the files must be 
        // opened and/or created), so we have to use try -- catch clauses to 
        // catch the exceptions.

        try {
           pw = new PrintWriter(new BufferedWriter(new FileWriter(file2)) );
              // this creates a PrintWriter attached to the physical file
              // file1 via FileWriter and BufferedWriter (unicode) character 
              // streams.  SUN advises putting the BufferedWriter in for 
              // efficiency when creating a PrintWriter stream, but the files 
              // we write here will be so small that we really don't need it in 
              // this case.
           
        }
        catch (SecurityException e) {
            printSecurityError();
            return false;  // writeOut operation failed
        }
        catch (IOException e) {
            System.out.println("Error opening input file.");
            return false;  // writeOut operation failed
        }

        // do the actual writing.  Note that the Member writeOut method does 
        // not throw an IO exception, but it does return true if all went well 
        // and false otherwise.
        i = 0;
        noError = true;
        while ( (i < L.length) && noError ) {
            noError = L[i].writeMember(pw);
            i++;
        }
        // don't need to flush stream because member writeOut method flushes 
        // output stream each time a record is written
        pw.close();  // close PrintWriter and output file

        if (!noError) {
           System.out.println("File output error occured");
        }

        return noError;

    }  // writeOutList

    public static void main(String[] args) {

       MailListSort mls = new MailListSort();

       // check for sane argument list
       if (args.length != 2) {
          mls.printUsageMessage();
          return;  // bail out of program
       }

       // read list in from file1 -- remember readInList returns true if it
       //   succeeds, false otherwise
       if (!mls.readInList(args[0])) {
          System.out.println("Ending program without sorting records");
          return;
       }

       // bubble sort list
       ObjectBubSort.bubSort(mls.L);

       // write list out to file2  -- remember writeOutList returns true if it
       //   succeeds, false otherwise
       if (!mls.writeOutList(args[1])) {
         System.out.println(
          "Ending program without completely writing out sorted member list");
          return;
       }

       // IT WORKED!  Give user a success message
       System.out.println("Sort operation succeeded.");
       System.out.println("Sorted member list contained in " + args[1]);

       return;

    } // main           

} // MailListSort
$DATA
JOAN
A
JONES
125 EAST MAIN
OGDEN
UTAH
82418
WILLIAM
R
SMITH
85 SOUTH 3RD
DEVILS SINKHOLE
TEXAS
79875
MARGARET
W
ANDERSON
1615 NORTH GREEN
URBANA
ILLINOIS
47815
JOAN
J
JONES
1341 SOUTH MONROE
CHICAGO
ILLINOIS
45217
MINNIE
D
MOUSE
133 INTERNATIONAL
DISNEYLAND
CALIFORNIA
98210
MICKEY
D
MOUSE
133 INTERNATIONAL
DISNEYLAND
CALIFORNIA
98210
MARGARET
W
ANDERSON
82 MESA AVE
EL PASO
TEXAS
79108
//
